import numpy as np
import os
import torch
from torch.utils.data import Dataset, DataLoader
import torch.nn as nn
import torch.optim as optim
from tqdm import tqdm
from segment_anything import sam_model_registry
import logging
import time
from pynvml import nvmlInit, nvmlDeviceGetHandleByIndex, nvmlDeviceGetMemoryInfo, nvmlDeviceGetPowerUsage, nvmlShutdown
import pickle
import random
# ============================ 记录代码运行的起始时间 ============================
start_time = time.time()
# ============================ 参数配置区域 ============================
params = {
    "npz_tr_path": ".../npz",  # 数据存放路径[todo]
    "model_save_path": ".../checkpoint",  # 模型存储路径[todo]
    "initial_checkpoint": ".../sam_vit_h_4b8939.pth",  # 初始权重文件路径[todo]
    "latest_model_name": "bestSAM.pth",  # 训练后保存的最优模型名称
    "predictions_pkl": "predictions.pkl",  # 推理预测结果文件名
    "ground_truths_pkl": "ground_truths.pkl",  # 推理GT结果文件名
    "gpu_metrics_pkl": "gpu_metrics.pkl",  # GPU性能指标文件名
}

# 模型和训练参数
config = {
    "model_type": "vit_h",  # 模型类型
    "device": "cuda:0",  # 使用的GPU设备
    "num_epochs": 100,  # 总训练epoch数
    "batch_size_train": 32,  # 训练批次大小（训练阶段分批次）
    "learning_rate": 1e-4,  # 学习率
    "weight_decay": 1e-5,  # 权重衰减
    "train_ratio": 0.8,  # 训练集占比
    "max_infer_samples": None,  # 最大推理样本数，None 表示使用所有推理数据集
}

# ============================ 日志设置 ============================
logging.basicConfig(filename="training_inference_log.log", level=logging.INFO, format="%(asctime)s - %(message)s")

# ============================ GPU 显存监控初始化 ============================
nvmlInit()
gpu_handle = nvmlDeviceGetHandleByIndex(0)  # 使用 GPU 0

# ============================ 自定义 Dataset ============================
class MyselfDataset(Dataset):
    def __init__(self, data_root, data_files):
        self.data_root = data_root
        self.data_files = data_files

    def __len__(self):
        return len(self.data_files)

    def __getitem__(self, index):
        file_path = os.path.join(self.data_root, self.data_files[index])
        with np.load(file_path) as data:  # 按需加载数据，并在读取后关闭文件
            img_embed = data['img_embeddings']
            gt2D = data['gts']
        return torch.tensor(img_embed).float(), torch.tensor(gt2D[None, :, :]).long()

# ============================ 数据划分 ============================
def split_data(npz_path, train_ratio):
    npz_files = sorted([f for f in os.listdir(npz_path) if f.endswith('.npz')])
    total_files = len(npz_files)
    train_size = int(total_files * train_ratio)
    train_files = npz_files[:train_size]
    infer_files = npz_files[train_size:]  # 使用剩余20%的数据作为推理集

    logging.info(f"Total .npz files found: {total_files}")
    logging.info(f"Training files: {len(train_files)}, Inference files: {len(infer_files)}")

    return train_files, infer_files

train_files, infer_files = split_data(params["npz_tr_path"], config["train_ratio"])

# ============================ 模型加载 ============================
sam_model = sam_model_registry[config["model_type"]](checkpoint=None).to(config["device"])
if os.path.exists(params["initial_checkpoint"]):
    logging.info(f"Loading initial weights from {params['initial_checkpoint']}")
    sam_model.load_state_dict(torch.load(params["initial_checkpoint"]))
else:
    raise FileNotFoundError(f"Initial checkpoint not found at {params['initial_checkpoint']}")
sam_model = torch.nn.DataParallel(sam_model, device_ids=[0])  # 使用 GPU 0 
sam_model = sam_model.to(config["device"])  # 将模型移动到设备
# ============================ 训练过程 ============================
sam_model.train()
optimizer = optim.Adam(sam_model.module.mask_decoder.parameters(), lr=config["learning_rate"], weight_decay=config["weight_decay"])
criterion = nn.BCEWithLogitsLoss()

train_dataset = MyselfDataset(params["npz_tr_path"], train_files)
train_dataloader = DataLoader(train_dataset, batch_size=config["batch_size_train"], shuffle=True)

best_loss = float('inf')

logging.info("Starting training...")
for epoch in range(config["num_epochs"]):
    epoch_loss = 0
    for step, (img_embedding, gt2D) in enumerate(tqdm(train_dataloader, desc=f"Training Epoch {epoch+1}/{config['num_epochs']}")):
        img_embedding, gt2D = img_embedding.to(config["device"]), gt2D.to(config["device"]).float()
        
        sparse_prompt_embeddings, dense_prompt_embeddings = sam_model.module.prompt_encoder(
            points=None,
            boxes=torch.tensor([1, 1, 224, 224]).unsqueeze(0).to(config["device"]),
            masks=None
        )
        
        mask_predictions, _ = sam_model.module.mask_decoder(
            image_embeddings=img_embedding,
            sparse_prompt_embeddings=sparse_prompt_embeddings,
            dense_prompt_embeddings=dense_prompt_embeddings,
            image_pe=sam_model.module.prompt_encoder.get_dense_pe(),
            multimask_output=False
        )
        
        loss = criterion(mask_predictions, gt2D)
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        epoch_loss += loss.item()

    epoch_loss /= len(train_dataloader)
    logging.info(f"Epoch {epoch+1}/{config['num_epochs']}, Loss: {epoch_loss:.4f}")
    print(f"Epoch {epoch+1}/{config['num_epochs']}, Loss: {epoch_loss:.4f}")

    if epoch_loss < best_loss:
        best_loss = epoch_loss
        torch.save(sam_model.state_dict(), os.path.join(params["model_save_path"], params["latest_model_name"]))

logging.info("Training completed.")

# ============================ 推理过程（逐样本） ============================
infer_dataset = MyselfDataset(params["npz_tr_path"], infer_files[:len(infer_files)])  # 推理只使用20%的数据
infer_dataloader = DataLoader(infer_dataset, batch_size=1, shuffle=False)  # 每次只推理一个样本

results = []
gpu_metrics = []  # 存储每个样本的GPU性能数据

# ============================ 推理性能指标 ============================
latencies_first_token = []  # 存储每个样本的首Token生成延时
latencies_all_tokens = []  # 存储每个Token的生成延时
throughput_per_token = []  # 每Token生成速度
step_times = []  # 总推理时间
mem_usages = []  # 显存占用
power_usages = []  # GPU功耗

logging.info("Starting inference...")
for step, (img_embedding, gt2D) in enumerate(tqdm(infer_dataloader, desc="Inference")):
    start_time = time.time()

    if config["max_infer_samples"] and step >= config["max_infer_samples"]:
        break
    
    with torch.no_grad():
        img_embedding = img_embedding.to(config["device"])

        sparse_prompt_embeddings, dense_prompt_embeddings = sam_model.module.prompt_encoder(
            points=None,
            boxes=torch.tensor([1, 1, 224, 224]).unsqueeze(0).to(config["device"]),
            masks=None
        )

        # 记录首Token生成延时
        token_start_time = time.time()
        
        mask_predictions, _ = sam_model.module.mask_decoder(
            image_embeddings=img_embedding,
            sparse_prompt_embeddings=sparse_prompt_embeddings,
            dense_prompt_embeddings=dense_prompt_embeddings,
            image_pe=sam_model.module.prompt_encoder.get_dense_pe(),
            multimask_output=False
        )
        
        # 记录生成所有Token的延时
        token_end_time = time.time()
        latencies_all_tokens.append(token_end_time - token_start_time)
        latencies_first_token.append(latencies_all_tokens[0])

        predictions = mask_predictions.cpu().numpy()
        results.append((predictions, gt2D.numpy()))

    # 计算每Token生成速度
    num_tokens = mask_predictions.numel()  # 获取Token数
    throughput = num_tokens / (token_end_time - token_start_time)
    throughput_per_token.append(throughput)
    
    # 记录GPU性能指标
    step_time = time.time() - start_time
    mem_info = nvmlDeviceGetMemoryInfo(gpu_handle)
    power_usage = nvmlDeviceGetPowerUsage(gpu_handle) / 1000  # W

    step_times.append(step_time)
    mem_usages.append(mem_info.used / 1024**2)
    power_usages.append(power_usage)

    logging.info(f"Sample {step+1}: Step Time: {step_time:.4f}s, Memory Used: {mem_info.used / 1024**2:.2f} MB, Power Usage: {power_usage:.2f} W")

logging.info("Inference completed.")
nvmlShutdown()

# ============================ 计算并记录推理性能指标 ============================
avg_first_token_latency = np.mean(latencies_first_token)
avg_latency_all_tokens = np.mean(latencies_all_tokens)
avg_throughput = np.mean(throughput_per_token)
avg_step_time = np.mean(step_times)
avg_mem_usage = np.mean(mem_usages)
avg_power_usage = np.mean(power_usages)

# 计算延时分布
tp50_latency = np.percentile(latencies_all_tokens, 50)
tp99_latency = np.percentile(latencies_all_tokens, 99)

logging.info(f"Average First Token Latency: {avg_first_token_latency:.4f}s")
logging.info(f"Average Token Latency: {avg_latency_all_tokens:.4f}s")
logging.info(f"TP50 Token Latency: {tp50_latency:.4f}s")
logging.info(f"TP99 Token Latency: {tp99_latency:.4f}s")
logging.info(f"Average Throughput per Token: {avg_throughput:.4f} tokens/s")
logging.info(f"Average Step Time: {avg_step_time:.4f}s")
logging.info(f"Average Memory Usage: {avg_mem_usage:.2f} MB")
logging.info(f"Average Power Usage: {avg_power_usage:.2f} W")

print(f"Average First Token Latency: {avg_first_token_latency:.4f}s")
print(f"Average Token Latency: {avg_latency_all_tokens:.4f}s")
print(f"TP50 Token Latency: {tp50_latency:.4f}s")
print(f"TP99 Token Latency: {tp99_latency:.4f}s")
print(f"Average Throughput per Token: {avg_throughput:.4f} tokens/s")
print(f"Average Step Time: {avg_step_time:.4f}s")
print(f"Average Memory Usage: {avg_mem_usage:.2f} MB")
print(f"Average Power Usage: {avg_power_usage:.2f} W")

# ============================ 保存推理结果 ============================
predictions_list = [item[0] for item in results]
gts_list = [item[1] for item in results]

with open(os.path.join(params["model_save_path"], params["predictions_pkl"]), 'wb') as f:
    pickle.dump(predictions_list, f)

with open(os.path.join(params["model_save_path"], params["ground_truths_pkl"]), 'wb') as f:
    pickle.dump(gts_list, f)

with open(os.path.join(params["model_save_path"], params["gpu_metrics_pkl"]), 'wb') as f:
    pickle.dump(gpu_metrics, f)

logging.info("Inference results and GPU metrics saved.")
print("Inference results saved as predictions.pkl, ground_truths.pkl, and gpu_metrics.pkl")
# ============================ 记录代码运行的结束时间并计算总耗时 ============================
end_time = time.time()  # 记录程序结束时间
total_runtime = end_time - start_time  # 计算总运行时间

logging.info(f"Total runtime: {total_runtime:.2f} seconds")
print(f"Total runtime: {total_runtime:.2f} seconds")
